#!/bin/env python3
from functools import lru_cache
from pydantic import TypeAdapter, ValidationError
from typing import Type, get_args, get_origin
import aiofiles
import asyncio
import logging
import os
import pathlib
import pprint
import psycopg
import psycopg_pool
import re
import urllib.parse

logger = logging.getLogger(__name__)


class c:
	HEADER = '\033[95m'
	BLUE = '\033[94m'
	CYAN = '\033[96m'
	GREEN = '\033[92m'
	RED = '\u001b[31m'
	ORANGE = '\u001b[38;5;202m'
	WARNING = '\033[93m'
	FAIL = '\033[91m'
	ENDC = '\033[0m'
	BOLD = '\033[1m'
	UNDERLINE = '\033[4m'

def stripes(cols=79, rows=3, on=3, off=3, on_char="#", off_char=" ", offset=0, sheer=1):
	out=""
	alphabet = on_char * on + off_char * off
	alphabet_len=len(alphabet)
	for row in range(0, rows):
		index = offset
		out += c.ORANGE
		for col in range(0, cols):
			alphabet_index = index % alphabet_len
			out += alphabet[alphabet_index]
			index += 1
		offset+=sheer
		out += c.ENDC +"\n"
	return out



def schema_to_compact(schema: dict) -> str:
	# Extract the title or name of the model
	model_name = schema.get('title', 'UnknownModel')
	
	# Initialize an empty list to hold field descriptions
	field_descriptions = []
	
	# Iterate over the properties in the schema to build field descriptions
	for field_name, details in schema.get('properties', {}).items():
		field_type = details.get('type', 'UnknownType')
		
		# For more complex types (e.g., arrays, nested models), you might need additional handling here
		if field_type == 'array':
			items = details.get('items')
			if isinstance(items, dict):
				item_type = items.get('type', 'UnknownType')
				field_type = f"List[{item_type}]"
		
		# Append the field description to the list
		field_descriptions.append(f"{field_name}:{field_type}")
	
	# Join all field descriptions into a single string
	fields_str = ', '.join(field_descriptions)
	
	# Return the compact representation of the model
	return f"{model_name}{{ {fields_str} }}"


def describe_type(input_type: Type) -> str:
	# Handle Pydantic models by checking for the .schema() method
	if hasattr(input_type, 'schema') and callable(getattr(input_type, 'schema')):
		schema = schema_to_compact(input_type.schema())
		return f"Pydantic {schema}"
	
	# Check for generic types (e.g., List[SomeType])
	origin_type = get_origin(input_type)
	if origin_type is not None:
		type_args = get_args(input_type)
		type_args_descriptions = [describe_type(t) for t in type_args]
		descs='\n, '.join(type_args_descriptions)
		return f"Generic type: {origin_type.__name__} of [{descs}]"
	# Fallback for standard or simple types
	return f"Type: {input_type}"


class Database:
	def __init__(self, db_uri, sql_dir):
		self.db_uri = db_uri
		self.query_dirs = set()
		self.queries = dict()
		self.prepared = dict()
		self.pool = None
		if sql_dir:
			self.query_dirs.add(sql_dir)

	async def preload_queries(self, path):
		"""Recursively load and register queries from .sql files in given path."""
		self.query_dirs.add(pathlib.Path(path))
		for dir in self.query_dirs:
			for sql_file in dir.rglob('*.sql'):
				# some/path/file.sql --> some.path.file
				query_key = sql_file.relative_to(dir).with_suffix('').as_posix().replace('/', '.')
				with open(sql_file, 'r', encoding='utf-8') as file:
					self.queries[query_key] = file.read()

	def reset_queries(self):
		"""Reset the internal cache of queries, forcing queries to be reloaded upon next invocation"""
		self.queries=dict()
		self.prepared=dict()

	async def load_query(self, query_key):
		"""Load and return a single query by first looking it up in the internal cache and falling back on looking for it on disk"""
		# Cache hit
		if query_key in self.queries:
			return self.queries[query_key]
		# Cache miss, load from disk
		try:
			# Make sure we know where to look
			if len(self.query_dirs) <= 0:
				raise Exception(f"No query dir(s) specified")
			# Convert the query_key to a file path
			for dir in self.query_dirs:
				file_path = dir / pathlib.Path(query_key.replace('.', '/') + '.sql')
				logger.info(f"file_path {file_path}")
					
				# Read the SQL query from file
				async with aiofiles.open(file_path, mode='r', encoding='utf-8') as file:
					sql_query = await file.read()
				# with open(file_path, 'r', encoding='utf-8') as file:
				#	sql_query = file.read()
				self.queries[query_key] = sql_query
				return sql_query
		except FileNotFoundError:
			logger.info(f"SQL file {os.path.abspath(file_path)}")
			logger.error(f"SQL file not found for key {query_key} at path {file_path}")
		except Exception as e:
			logger.error(f"{type(e).__name__}: Error loading SQL file for key {query_key}: {e}")
		return None

	def _get_pool(self):
		self.pool = self.pool or psycopg_pool.AsyncConnectionPool(self.db_uri) # , min_size=min_conn, max_size=max_conn
		return self.pool


	def _adapt(self, raw, item_type, do_debug=False):
		if not item_type:
			return raw
		if do_debug:
			logger.info(f"Adapting {pprint.pformat(raw)} to {pprint.pformat(item_type)}")
		try:
			return TypeAdapter(item_type).validate_python(raw)
		except Exception as e:
			logger.warning(f"Adaption error details")
			logger.warning(f"                       raw: {pprint.pformat(raw)}")
			logger.warning(f"                 type(raw): {type(raw)}")
			logger.warning(f"                 item_type: {item_type}")
			logger.warning(f"  describe_type(item_type): {describe_type(item_type)}")
			raise


	async def direct_query(self, query, query_name=None, params=dict(), mode="none", item_type=None, prepare=True, do_debug=True):
		cursor = None
		try:
			if not query:
				raise Exception("No query")
			# Make sure we actually have a pool
			pool = self._get_pool()
			if not pool:
				raise Exception("No pool")
			# TODO: Look into implementing pipeline mode
			# See: https://www.psycopg.org/psycopg3/docs/advanced/pipeline.html
			# Acquire a connection from the pool
			async with pool.connection() as connection:
				async with connection.transaction() as transaction:
					async with connection.cursor(row_factory=psycopg.rows.dict_row) as cursor:
						# See https://www.psycopg.org/psycopg3/docs/advanced/prepare.html
						if do_debug:
							logger.info(f"Preparing query '{query_name}'")
							logger.info("\n\n" + query+"\n\n")
							logger.info(f"With data")
							logger.info("\n\n" + pprint.pformat(params)+"\n\n")
						await cursor.execute(query = query, params = params, prepare = prepare)
						if mode == "many":
							res = self._adapt(await cursor.fetchall(), item_type, do_debug)
							if do_debug:
								logger.info(f"Returning many data: '{pprint.pformat(res)}'")
							return res, None
						elif mode == "one":
							res = self._adapt(await cursor.fetchone(), item_type, do_debug)
							if do_debug:
								logger.info(f"Returning one data: '{pprint.pformat(res)}'")
							return res, None
						elif mode == "exists":
							res = await cursor.fetchone()
							if item_type:
								logger.warning(f"item_type speficied '{item_type}' for query with mode exists")
							if do_debug:
								logger.info(f"Returning exist data: '{pprint.pformat(res)}'")
							return res is not None, None
						elif mode == "none":
							if item_type:
								logger.warning(f"item_type speficied '{item_type}' for query with mode none")
							if do_debug:
								logger.info(f"Returning for none")
							return None, None
						else:
							raise Exception(f"Unknown mode {mode}. Should be one of [many, one, exists, none]")
		except Exception as e:
			logger.error("\n\n"+stripes())
			logger.error("")
			logger.error("")
			logger.error(f"###############################################")
			logger.error(f"+       Error: ({type(e).__name__})")
			logger.error(f"#    Querying: '{query_name}'\n\n{query}\n")
			if cursor and query and "cursor" in query:
				logger.error(f"#    Injected: '{cursor.query.decode()}'")
			logger.error(f"#         On")
			logger.error(f"#         + db_uri =   '{self.db_uri}'")
			logger.error(f"#         For")
			logger.error(f"#         + mode =   '{mode}'")
			logger.error(f"#         + params = '{pprint.pformat(params)}'")
			logger.error(f"# Failed with:")
			if isinstance(e,psycopg.Error):
				logger.error(f"# + e.type:     '{type(e).__name__}'")
				logger.error(f"# + e.sqlstate: '{e.sqlstate}'")
				if getattr(e, "pgconn", None):
					logger.error(f"# + e.pgconn:   '{e.pgconn}'")
				if getattr(e, "pgresult", None):
					logger.error(f"# + e.pgresult: '{e.pgresult}'")
				if e.diag:
					logger.error(f"# + e.diag: '{e.diag}'")
					error_attrs = [
						"column_name", "constraint_name", "context", "datatype_name",
						"internal_position", "internal_query", "message_detail", "message_hint",
						"message_primary", "schema_name", "severity", "source_file",
						"source_function", "source_line", "sqlstate", "statement_position",
						"table_name", "severity_nonlocalized"
					]
					for attr in error_attrs:
						value = getattr(e.diag, attr, None)  # Use getattr to safely handle missing attributes
						if value is not None:  # Only log if the attribute is present and not None
							logger.error(f"#   + {attr}: {value}")
			elif isinstance(e, ValidationError):
				logger.error(f"Validation error: with item_type '{item_type}'")
			else:
				# Unexpected errors
				logger.error(f"Unexpected error: {type(e).__name__}")
			logger.error(f"#  Stacktrace: \n\n{e}\n")
			logger.error("\n\n"+stripes())
			return None, str(e)


	async def key_query(self, query_key, params=dict(), mode="none", item_type=None, prepare=True, do_debug=False):
		query = await self.load_query(query_key)
		if query:
			logger.info(f"key_query params='{pprint.pformat(params)}' prepare={prepare}")
			return await self.direct_query(query = query, query_name=query_key, params = params, mode = mode, item_type=item_type, prepare=prepare, do_debug=do_debug)
		return None, "No query"

	async def query_many(self, query_key, params=dict(), item_type=None, prepare=True, do_debug=False):
		if do_debug:
			logger.info(f"query_many params='{pprint.pformat(params)}' prepare={prepare}")
		return await self.key_query(query_key = query_key, params = params, mode = "many", item_type=item_type, prepare=prepare, do_debug=do_debug)

	async def query_one(self, query_key, params=dict(), item_type=None, prepare=True, do_debug=False):
		if do_debug:
			logger.info(f"query_one params='{pprint.pformat(params)}' prepare={prepare}", prepare=True, do_debug=False)
		return await self.key_query(query_key = query_key, params = params, mode = "one", item_type=item_type, prepare=prepare, do_debug=do_debug)

	async def query_none(self, query_key, params=dict(), prepare=True, do_debug=False):
		if do_debug:
			logger.info(f"query_none params='{pprint.pformat(params)}' prepare={prepare}", prepare=True, do_debug=False)
		return await self.key_query(query_key = query_key, params = params, mode = "none", prepare=prepare, do_debug=do_debug)

	async def query_exists(self, query_key, params=dict(), prepare=True, do_debug=False):
		if do_debug:
			logger.info(f"query_exists params='{pprint.pformat(params)}' prepare={prepare}")
		return await self.key_query(query_key = query_key, params = params, mode = "exists", prepare=prepare, do_debug=do_debug)


	async def execute_old(self, query_key, **params):
		"""Execute a registered query by its key."""
		if query_key not in self.queries:
			raise ValueError(f"Query '{query_key}' not found.")
		
		async with self.pool.connection() as connection:
			async with connection.cursor() as cur:
				await cur.execute(self.queries[query_key], params)
				# Assuming a fetch is needed, adjust as necessary
				return await cur.fetchall()


def _q(s):
	if not s:
		return ""
	return urllib.parse.quote(s)

def db_uri_from_conf(config:dict) ->str :
	return f"postgresql://{_q(config.get('db-username'))}:{_q(config.get('db-password'))}@{config.get('db-hostname')}:{config.get('db-port')}/{_q(config.get('db-database'))}"

#@lru_cache
def get_database(config:dict) -> Database:
	uri = db_uri_from_conf(config)
	logger.info(f"Trying to connect to database with '{uri}'")
	db = Database(db_uri=uri, sql_dir=config.get("db-sql-dir"))
	return db




from fastapi import Depends
from octomy.batch.db import Database
from octomy.batch.types import *
from octomy.utils import human_delta
from pathlib import Path
import datetime
import octomy.utils
import octomy.utils.Watchdog
import glob
import importlib
import importlib.util
import json
import logging
import os
import pprint
import pydantic.version
import random
import re
import string
import sys
import time
import traceback
from .types import *

logger = logging.getLogger(__name__)


CLEANUP_INTERVAL_MS = 0.999


def log_item(item):
	# logger.info(pprint.pformat(item))
	logger.info(item_to_str(item))



def item_to_str(item):
	prefix = "#=- "
	ret = "\n"
	if not item:
		ret = f"""
{prefix}
{prefix}Job:    NONE
{prefix}
"""
		return ret
	ret += f"{prefix}\n"
	type = item.get("type", "unknown-type")
	id = item.get("id", "XXXXXXX")
	ret += f"{prefix}BATCH JOB {type}: {id}\n"
	try:
		now = datetime.now()
		created_ago = human_delta(now - item.get("created_at"), None)
		updated_ago = human_delta(now - item.get("updated_at"), None)
		ret += f"{prefix}Created: {created_ago}, Updated: {updated_ago} ####\n"
	except:
		pass
	try:
		source = item.get("source")
		if source:
			ret += f"{prefix}Source: {source}\n"
		status = item.get("status")
		if status:
			ret += f"{prefix}Status: {status}\n"
	except:
		pass
	data_raw = item.get("data")
	if data_raw:
		ret += f"{prefix}Data:\n\n"
		try:
			data = json.loads(data_raw)
			data_str = json.dumps(data, indent=3, sort_keys=True, default=str)
			ret += data_str + "\n\n"
		except json.JSONDecodeError as e:
			ret += f"{prefix}JSON PARSE ERROR\n"
	result_raw = item.get("result")
	if result_raw:
		ret += f"{prefix}Result:\n\n"
		try:
			result = json.loads(result_raw)
			result_str = json.dumps(result, indent=3, sort_keys=True, default=str)
			ret += result_str + "\n\n"
		except json.JSONDecodeError as e:
			ret += result_raw + "\n\n"
	ret += f"{prefix}\n"
	return ret


def random_token(length=10):
	alphabet = string.ascii_letters + string.digits
	return "".join(random.choice(alphabet) for i in range(length))



class Processor:
	def __init__(self, config, dbc, do_debug=False):
		self.do_debug = do_debug
		self.worker_id = None
		self.config = config
		self.dbc = dbc
		self.db = octomy.batch.db.Database(config, dbc)
		self.do_log = self.config.get("batch-log-logging", True)
		self.last_cleanup_time = datetime.datetime.now()
		self.task_filter = re.compile("^([0-9a-z\-_]+(\/[0-9a-z\-_]+)*)$")
		self.last_status_time = None
		self.callables = {}
		self.callables_params = {}
		self.entry_name = "batch_filter_entrypoint"
		self.param_name = "batch_filter_parameters"

	async def get_id(self):
		if not self.worker_id:
			self.worker_id = random_token(10)
			logger.info(f"Batch Processor {self.worker_id} instanciated")
		return self.worker_id

	def __del__(self):
		if self.worker_id:
			logger.info(f"Batch Processor {self.worker_id} deleted")
			self.worker_id = None


	async def create_tables(self):
		await self.dbc.query_none("overwatch.batch.create_batch_status_enum", prepare=False)
		await self.dbc.query_none("overwatch.batch.create_batch_items", prepare=False)
		await self.dbc.query_none("overwatch.batch.create_gcra_throttle", prepare=False)

	async def verify(self):
		modules = self.list_modules()
		ok = []
		failed = []
		for (name, error) in modules:
			if error:
				failed.append((name, error))
			else:
				ok.append(name)
		if ok:
			logger.info("Available modules:")
			for name in ok:
				logger.info(f" + '{name}'")
		if failed:
			message = ""
			logger.info("Failed modules:")
			for (name, error) in failed:
				logger.error(f" + '{name}': {error}")
				message += f"{name}: {error}\n"
			return False, message
		logger.info(f"Extended logging={self.do_log}")
		return True, ""

	async def load_module_by_filepath(self, module_name, module_filepath):
		module = None
		failure = None
		spec = None
		try:
			spec = importlib.util.spec_from_file_location(module_name, module_filepath)
			spec.submodule_search_locations = list(__import__(__name__).__path__)
			module = importlib.util.module_from_spec(spec)
			spec.loader.exec_module(module)
		except Exception as e:
			failure = f"Import of module '{module_name}'from file '{module_filepath}' had errors ({e}). (Spec was {pprint.pformat(spec)})"
			module = None
		return module, failure

	async def load_module_by_package(self, module_name, package_path=None):
		module = None
		failure = None
		try:
			module = importlib.import_module(module_name, package_path)
		except Exception as e:
			failure = f"Import of module '{module_name}'from package '{package_path}' had errors ({e})"
			module = None
		return module, failure

	async def get_item_types(self):
		module_root_raw = self.config.get("batch-filter-root", "/tmp/inexistant")
		module_root_object = Path(module_root_raw).resolve(strict=False)
		module_root = module_root_object.as_posix()
		if not module_root:
			logger.warning("Module root was not set")
		if not os.path.exists(module_root):
			logger.warning(f"Module root '{module_root}' did not exist")
		logger.info(f"Looking for types in '{module_root}':")
		ret = {}
		for path_object in module_root_object.rglob("*.py"):
			path_object = path_object.resolve(strict=False)
			path = path_object.as_posix()
			if path_object.name.startswith("__"):
				logger.warning(f"Skipping invalid path {path}")
				continue
			if not path.startswith(module_root):
				logger.warning(f"Skipping invalid path {path}")
				continue
			path_difference = path[len(module_root) + 1 :]
			name = Path(path_difference).with_suffix("").as_posix()
			ret[name] = path
			logger.info(f"  Found {name} (from {path})")
		return ret

	async def list_modules(self):
		module_root_raw = self.config.get("batch-filter-root", "/tmp/inexistant")
		module_root_object = Path(module_root_raw).resolve(strict=False)
		module_root = module_root_object.as_posix()

		ret = []
		failure = None
		if not module_root:
			ret.append(("module_root", f"Module root was not set"))
		elif not os.path.exists(module_root):
			ret.append(("module_root", f"Module root '{module_root}' did not exist"))
		else:
			files_objects = module_root_object.rglob("*.py")
			for file_object in files_objects:
				module_filepath = file_object.as_posix()
				if not module_filepath:
					logger.warn(f"Not posix path: {file_object}")
					continue
				if file_object.name.startswith("__"):
					logger.warn(f"Skipping internal: {file_object}")
					continue
				if octomy.utils.file_contains_str(module_filepath, self.entry_name):
					try:
						module, failure = self.load_module_by_filepath(module_filepath, module_filepath)
						if module:
							if hasattr(module, self.entry_name):
								entry_method = getattr(module, self.entry_name)
								if not callable(entry_method):
									entry_method = None
									failure = f"Entrypoint was not callable in filter module {module_filepath}"
							else:
								failure = f"Filter module {module_filepath} did not have method {self.entry_name}"
						else:
							failure = f"Filter module {module_filepath} could not be loaded because: {failure}"
					except Exception as e:
						failure = f"Import of module '{module_filepath}' had errors and was skipped ({e})"
				else:
					failure = f"No entrypoint found in filter module {module_filepath}"
				# logger.warn(f"Appending stat: {module_filepath}:{failure}")
				ret.append((module_filepath, failure))
		return ret

	async def get_callable_for_type_raw(self, t):
		match = self.task_filter.match(t)
		module_name = None
		entry_method = None
		param_dict = None
		failure = None
		# Is this a match?
		if match:
			# Exctract the data we want
			module_name = match.group(1)
			module_filename = module_name + ".py"
			module_filepath = os.path.join(self.config.get("batch-filter-root", "/dev/null"), module_filename)
			module = None
			if os.path.exists(module_filepath):
				if octomy.utils.file_contains_str(module_filepath, self.entry_name):
					try:
						module, failure = self.load_module_by_filepath(module_name, module_filepath)
						if module:
							if hasattr(module, self.entry_name):
								entry_method = getattr(module, self.entry_name)
								if not callable(entry_method):
									entry_method = None
									failure = f"Entrypoint was not callable in filter module {module_filepath}"
								if hasattr(module, self.param_name):
									param_dict = getattr(module, self.param_name)
							else:
								failure = f"Filter module {module_filepath} did not have method {self.entry_name}"
						else:
							failure = f"Filter module {module_filepath} could not be loaded because: {failure}"
					except Exception as e:
						failure = f"Import of module '{module_name}' had errors and was skipped ({e})"
				else:
					failure = f"No entrypoint found in filter module {module_filepath}"
			else:
				failure = f"Could not find filter module {module_filepath}"
		else:
			failure = f"No match for type {t}"
		return entry_method, param_dict, failure

	async def get_callable_for_type(self, t):
		if t in self.callables:
			return self.callables.get(t), self.callables_params.get(t), None
		else:
			cb, params, failure = self.get_callable_for_type_raw(t)
			if cb:
				self.callables[t] = cb
			if params:
				self.callables_params[t] = params
			return cb, params, failure

	async def _watchdog_hang_detected(self):
		logger.error("Watchdog triggered exception, restarting batch")
		sys.exit(99)

	async def _execute_safely(self, entrypoint, params, item):
		# logger.info("SAFE EXECUTION STARTED!")
		failure = None
		result = None
		watchdog = None
		try:
			# We cap runtime at TTL from which ever is set first; item ttl, param ttl or environment fallback value
			timeout = self.config.get("batch-default-ttl", 0)
			timeout = params.get("ttl_seconds", timeout) if params else timeout
			timeout = item.get("ttl_seconds", timeout) if item else timeout
			if timeout and timeout > 0:
				logger.info(f"Using watchdog with ttl={timeout}sec")
				watchdog = octomy.utils.Watchdog.Watchdog(timeout, self._watchdog_hang_detected)
			else:
				logger.info(f"No watchdog enabled")
			# logger.info("££££ Entry")
			unpackable = entrypoint(item, self.config)
			try:
				result, failure = unpackable
			except TypeError as te:
				failure = f"Filter did not return exactly one value and one error"
				logger.warning(failure)
				return None, failure
			# logger.info("££££ Exit")
		except octomy.utils.Watchdog.Watchdog:
			logger.warning("Watchdog triggered exception")
			failure = f"Execution timed out after {timeout} seconds"
		except Exception as e:
			logger.warning("")
			logger.warning("")
			logger.warning(f"###############################################")
			logger.warning(f"#    Batch item: {item} on worker {self.worker_id}")
			logger.warning(f"# Failed with: {e} ({type(e).__name__}, {e.args})")
			failure = f"{e}  ({type(e).__name__}, {e.args})"
			logger.warning(f"#          At:")
			traceback.print_exc()
			logger.warning(f"#######################################")
			logger.warning("")
			logger.warning("")
		if None is not watchdog:
			watchdog.stop()
		# logger.info("SAFE EXECUTION FINISHED!")

		watchdog = None
		return result, failure

	async def execute_one_batch_item(self):
		"""
		Take ownership of one batch item and make sure it is properly executed and updated with status underway
		"""
		worker_id = self.get_id()
		updated_status = {"from_status": JobStatusEnum.PENDING, "to_status": JobStatusEnum.IN_PROGRESS, "in_progress_status": JobStatusEnum.IN_PROGRESS, "worker_id": worker_id}
		item, err = await self.dbc.key_query(query_key = "overwatch.batch.book_batch_item", params=updated_status, mode = "one", item_type = dict)
		if item:
			if self.do_log:
				logger.info("Processing item:")
				log_item(item)
			id = item.get("id", None)
			t = item.get("type", None)
			updated_at = item.get("updated_at", None)
			if id and t and updated_at:
				entrypoint, params, failure = self.get_callable_for_type(t)
				if entrypoint and failure:
					entrypoint = None
					failure = f"{failure} AND ENTRYPOINT WAS SET!"
				result = None
				if entrypoint:
					result, failure = self._execute_safely(entrypoint, params, item)
				if failure:
					logger.warning("¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤ BATCH FILTER FAILED WITH ¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤")
					logger.warning(failure)
					logger.warning("")
				(id2, updated_at2), err2 = await self.db.bump_batch_item(id=id, status=JobStatusEnum.FAILED if failure else JobStatusEnum.DONE, error=failure, updated_at=updated_at, result=result) or (None, None)
				if self.do_log:
					logger.info(f"id2={id2}, updated_at2={updated_at2}, id={id}, updated_at={updated_at}, {err2}")
				return True
			else:
				logger.warning(f"Missing data for item id={id}, type={t}, updated_at={updated_at}")
				logger.error(f"err={err}")
		# else:
		# logger.info(f"No pending items found")
		return False

	async def execute_one_throttled_batch_item(self):
		"""
		Take ownership of one batch item and make sure it is properly executed and updated with status underway
		NOTE: Throttled version
		"""
		worker_id = self.get_id()
		updated_status = {"from_status": JobStatusEnum.PENDING, "to_status": JobStatusEnum.IN_PROGRESS, "in_progress_status": JobStatusEnum.IN_PROGRESS, "worker_id": worker_id}
		item, err = await self.dbc.key_query(query_key = "overwatch.batch.book_throttled_batch_item", params=updated_status, mode = "one", item_type = dict)
		if item:
			if self.do_log:
				logger.info("Processing throttled item:")
				log_item(item)
			id = item.get("id", None)
			t = item.get("type", None)
			updated_at = item.get("updated_at", None)
			if id and t and updated_at:
				if item.get("wait_millis", 0) > 0:
					logger.warning("Item not booked")
					logger.warning(pprint.pformat(item))
					failure = "Not booked"
				else:
					entrypoint, params, failure = self.get_callable_for_type(t)
					if entrypoint and failure:
						entrypoint = None
						failure = f"{failure} AND ENTRYPOINT WAS SET!"
					result = None
					if entrypoint:
						result, failure = self._execute_safely(entrypoint, params, item)
					if failure:
						logger.warning("¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤ BATCH FILTER FAILED WITH ¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤")
						logger.warning(failure)
						logger.warning("")
				id2, updated_at2 = await self.db.bump_batch_item(id=id, status=JobStatusEnum.FAILED if failure else JobStatusEnum.DONE, error=failure, updated_at=updated_at, result=result) or (None, None)
				if self.do_log:
					logger.info(f"id2={id2}, updated_at2={updated_at2}, id={id}, updated_at={updated_at}")
				return True
			else:
				logger.warning(f"Missing data for item id={id}, type={t}, updated_at={updated_at}")
				logger.error(f"err={err}")
		# else:
		# logger.info(f"No pending items found")
		return False

	# Make sure data is a string ready for insertion into db
	async def _prepare_data(self, data):
		if data == None:
			return None
		data_type = type(data)
		if data_type is dict:
			data = json.dumps(data, indent=3, sort_keys=True, default=str)
		elif data_type is not str:
			data = str(data)
		return data

	async def insert_batch_item(self, type="test", data=None, ttl_seconds=None, result=None, priority=50, source=None, throttle_key=None, throttle_limit=1, throttle_period=1):
		"""
		Insert a new batch item into the database, ready for execution
		"""
		data = self._prepare_data(data)
		# fmt: off
		return await self.db.insert_batch_item(
			{
				"priority": priority,
				"ttl_seconds": ttl_seconds, 
				"data": data,
				"result": result,
				"type": type,
				"status": JobStatusEnum.PENDING,
				"throttle_key": throttle_key,
				"throttle_limit": throttle_limit,
				"throttle_period": throttle_period,
				"source": source
			}
		)
		# fmt: on

	async def _update_hang_status(self):
		updated_status= {"from_status": JobStatusEnum.IN_PROGRESS, "to_status": JobStatusEnum.HUNG, "in_progress_status": JobStatusEnum.IN_PROGRESS}
		db_items, db_err = await self.dbc.key_query(query_key = "overwatch.batch.update_hang_status", params=updated_status, mode = "one", item_type = dict)
		if db_err:
			logger.warning(f"Could not update hang status: {db_err}")
		return db_items

		return await self.db.update_batch_hang_status()

	async def retry_hung_jobs(self):
		updated_status = {"from_status": JobStatusEnum.IN_PROGRESS, "to_status": JobStatusEnum.PENDING, "in_progress_status": JobStatusEnum.IN_PROGRESS}
		db_items, db_err = await self.dbc.key_query(query_key = "overwatch.batch.bump_batch_items", params=updated_status, mode = "one", item_type = dict)
		if db_err:
			logger.warning(f"Could retry hung jobs: {db_err}")
		return db_items
		return await self.db.bump_batch_items(JobStatusEnum.HUNG, JobStatusEnum.PENDING)


	async def delete_jobs_with_status(self, status):
		query = {"status": status}
		db_items, db_err = await self.dbc.key_query(query_key = "overwatch.batch.delete_batch_items_by_status", params = query, mode = "none", do_debug = self.do_debug)
		if db_err:
			logger.warning(f"Could not delete '{query.get('status')}' jobs: {db_err}")
		return db_items


	async def delete_all_jobs(self):
		query = {}
		db_items, db_err = await self.dbc.key_query(query_key = "overwatch.batch.delete_all", params = query, mode = "none", do_debug = self.do_debug)
		if db_err:
			logger.warning(f"Could not delete all {db_err}")
		return db_items


	async def bump_job_by_id(self, id, status):
		return await self.db.bump_batch_item_status(id, status)


	async def retry_job_by_id(self, id):
		return await self.db.bump_batch_item_status(id, JobStatusEnum.PENDING)


	async def delete_job_by_id(self, id):
		query = {"id": id}
		db_items, db_err = await self.dbc.key_query(query_key = "overwatch.batch.delete_by_id", params = query, mode = "none", do_debug = self.do_debug)
		if db_err:
			logger.warning(f"Could not delete JOB BY ID '{query.get('ID')}': {db_err}")
		return db_items


	async def get_job_counts(self):
		#raw, err = await self.db.get_job_counts()
		raw, err = await self.dbc.key_query(query_key="overwatch.batch.get_job_counts", params={}, mode = "many", item_type = List[JobCount])
		if raw:
			out = {}
			totals = {}
			for row in raw:
				t = row.type
				s = row.status
				c = row.count
				totals[s] = totals.get(s, 0) + c
				out[t] = out.get(t, {})
				out[t][s] = c
			out["total"] = totals
			return out, None
		else:
			logger.error(f"err={err}")
			return None, err

	async def get_job_times(self):
		raw = await self.db.get_job_times()
		return raw

	async def get_worker_stats(self):
		raw, err = await self.db.get_worker_stats()
		return raw, err

	async def get_status(self):
		status, status_err = await self.dbc.key_query(query_key="overwatch.batch.get_status_counts", params={}, mode = "many", item_type = List[StatusCount])
		counts, counts_err = await self.get_job_counts()
		times, times_err = await self.dbc.key_query(query_key="overwatch.batch.get_job_times", params={"statuses": ["done"]}, mode = "many", item_type = List[JobTimes])
		workers, workers_err = await self.get_worker_stats()
		# fmt: off
		status = {
			  "status": status
			, "counts": counts
			, "times": times
			, "workers": workers
			, "errors": {
				  "status": status_err
				, "counts": counts_err
				, "times": times_err
				, "workers": workers_err
			}
		}
		# fmt: on
		if any([status_err, counts_err, times_err, workers_err]):
			return None, f"status_err={status_err}, counts_err={counts_err}, times_err={times_err}, workers_err={workers_err}"
		return status, None

	async def process(self):
		# Perform some cleanup on occation
		now = datetime.datetime.now()
		if now - self.last_cleanup_time > datetime.timedelta(seconds=CLEANUP_INTERVAL_MS):
			self.last_cleanup_time = now
			self._update_hang_status()
		# Try to process one item
		if not self.execute_one_throttled_batch_item():
			# Lets not get stuck in a loop!
			time.sleep(1)



from dateutil.relativedelta import relativedelta
from urllib.parse import urlparse
from uuid import UUID
import asyncio
import base64
import datetime
import hashlib
import inspect
import json
import logging
import mmap
import os
import pkg_resources
import random
import requests
import sys
import time


logger = logging.getLogger(__name__)

# We want 4 second timeout for HTTP requests
HTTP_TIMEOUT = 4.0
_requests_session_instance = None


def redact(val, minimum=20):
	total = len(val)
	scramble_n = max(minimum, total - 3)
	show_n = max(min(total - minimum, 3), 0)
	out = val[0:show_n] + "*" * (scramble_n)
	return out


def file_contains_str(filename, str):
	return file_contains_bytes(filename, str.encode("utf-8"))


def file_contains_bytes(filename, bytes):
	# mmap does not work for empty files
	if os.stat(filename).st_size == 0:
		return False
	with open(filename, "rb", 0) as file, mmap.mmap(file.fileno(), 0, access=mmap.ACCESS_READ) as s:
		if s.find(bytes) != -1:
			return True
	return False


# Merge two dictionaries
def merge(dict1, dict2):
	res = {**dict1, **dict2}
	return res


# Determine if given URL is valid
def is_url_valid(url):
	try:
		result = urlparse(url)
		# 		if not result.scheme:
		# 			print ("No scheme")
		# 		if not result.netloc:
		# 			print ("No netloc")
		# 		if not result.path:
		# 			print ("No path")
		# 		if not result.scheme in ['http', 'https']:
		# 			print ("Scheme invalid")

		return all([result.scheme, result.netloc]) and result.scheme in ["http", "https"]
		# result.netloc,
	except:
		return False


def is_valid_uuid(uuid_to_test, version=4):
	"""
	Check if uuid_to_test is a valid UUID.

	Parameters
	----------
	uuid_to_test : str
	version : {1, 2, 3, 4}

	Returns
	-------
	`True` if uuid_to_test is a valid UUID, otherwise `False`.

	Examples
	--------
	>>> is_valid_uuid('c9bf9e57-1685-4c89-bafb-ff5af830be8a')
	True
	>>> is_valid_uuid('c9bf9e58')
	False
	"""
	try:
		uuid_obj = UUID(uuid_to_test, version=version)
	except ValueError:
		return False

	return str(uuid_obj) == uuid_to_test


# Create a hash hex digest from string
def hashstr(raw):
	if type(raw) != type(""):
		raw = f"{raw}"
	sha256 = hashlib.sha256()
	sha256.update(raw.encode("utf-8"))
	return sha256.hexdigest().upper()


def hashfile(fn):
	file = open(fn, "r+")
	size = os.path.getsize(fn)
	sha256 = hashlib.sha256()
	sha256.update(mmap.mmap(file.fileno(), size))
	return sha256.update(mmap.mmap(file.fileno(), size))


# Turn invalid string into naive URL by prepending schema and appending path
def decorate_url(url):
	if not is_url_valid(url):
		url = "http://{url}/".format(url=url)
	return url


# Return dict with url split into components as appropriate for storing in database
def split_url(url_full):
	ret = {"url_hostname": "", "url_schema": "", "url_path": "", "url_query": "", "url_fragment": ""}
	try:
		parts = urlparse(url_full)
		if parts.netloc is not None:
			ret["url_hostname"] = parts.netloc
		if parts.scheme is not None:
			ret["url_schema"] = parts.scheme
		if parts.path is not None:
			ret["url_path"] = parts.path
		if parts.query is not None:
			ret["url_query"] = parts.query
		if parts.port is not None:
			ret["url_port"] = parts.port
		if parts.fragment is not None:
			ret["url_fragment"] = parts.fragment
	except Exception as e:
		# print(f"ERROR SPLITTING URL: {e}")
		pass
	return ret


def flatten_headers(headers):
	out = ""
	for k, v in headers.items():
		out += f"'{k}'='{v}'\n"
	return out


def verify_input_files(input_options):
	input_files = []
	notes = []
	if input_options:
		for file in input_options:
			note = "[OK]"
			if not os.path.exists(file):
				note = "[MISSING, Skipped]"
			elif not os.path.isfile(file):
				note = "[NOT A FILE, Skipped]"
			elif os.access(file, os.R_OK):
				note = "[NOT READABLE, Skipped]"
			else:
				input_files.append(file)
			notes.append(f' "{file}"   {note}')
		return input_files, notes
	else:
		return None, None


def verify_output_file(output_file):
	valid_output_file = None
	if output_file:
		output_path = os.path.dirname(os.path.realpath(output_file))
		os.makedirs(output_path, exist_ok=True)
		note = "[OK]"
		if not os.path.exists(output_path):
			note = "[PATH MISSING]"
		elif os.access(output_file, os.W_OK):
			note = "[NOT WRITABLE]"
		else:
			valid_output_file = output_file
	return valid_output_file, note


# Look at result from scrape and produce a dictionary with relevant data or error output
def get_data_for_scrape_result(result):
	http_status_code = result.get("http_status_code", 500)
	reason = result.get("reason", "")
	print(f"SCRAPING COMPLETED {http_status_code}, {reason}")

	data = None
	error = ""
	ok = True
	if http_status_code != 200:
		error = f"http code {http_status_code} with reason {reason}"
		ok = False
	else:
		headers = result.get("headers", "")
		page_source = result.get("page_source", "")
		if not page_source:
			error = "no page_source"
			ok = False
		else:
			data = None
			try:
				data = json.loads(page_source)
			except Exception as e:
				error = f"json parse failed for {page_source[:100]} with {e}"
				ok = False
			if ok:
				t = __builtins__.type(data)
				if not data:
					error = f"json parse returned no data"
					ok = False
				# elif t is not '<class \'dict\'>':
				# 	error=f"json parse returned not dict ({t})"
				# 	ok=False
	return data, ok, error


attrs = ["millenia", "centuries", "decades", "years", "months", "days", "hours", "minutes", "seconds"]

human_readable = lambda delta: ["%d %s" % (getattr(delta, attr), getattr(delta, attr) > 1 and attr or attr[:-1]) for attr in attrs if getattr(delta, attr)]


def human_delta(td_object: datetime.timedelta, max: int = 0):
	ms = int(td_object.total_seconds() * 1000)
	if ms == 0:
		return "0 ms"
	sign = ""
	if ms < 0:
		ms = -ms
		sign = "-"
	# fmt: off
	periods = [
		("year",  1000 * 60 * 60 * 24 * 365),
		("month", 1000 * 60 * 60 * 24 * 30),
		("day",   1000 * 60 * 60 * 24),
		("hr",    1000 * 60 * 60),
		("min",   1000 * 60),
		("sec",   1000),
		("ms", 1)
	]
	# fmt: on

	strings = []
	ct: int = 0
	for period_name, period_ms in periods:
		if ms > period_ms:
			period_value, ms = divmod(ms, period_ms)
			# has_s = "s" if period_value > 1 else ""
			# strings.append("%s %s%s" % (period_value, period_name, has_s))
			strings.append(f"{period_value} {period_name}")
			ct += 1
			if max > 0 and ct > max:
				break
	return sign + ", ".join(strings)  # + f"({td_object}, {ms})"


def human_bytesize(bytes: int, max: int = 0):
	if bytes == 0:
		return "0 bytes"
	if bytes < 0:
		return "negative bytes (error)"
	# fmt: off
	KiB=1024
	periods = [
		("PiB", KiB*KiB*KiB*KiB*KiB),
		("TiB", KiB*KiB*KiB*KiB),
		("GiB", KiB*KiB*KiB),
		("MiB", KiB*KiB),
		("KiB", KiB),
		("bytes", 1)
	]
	# fmt: on

	strings = []
	ct: int = 0
	for period_name, period_bytes in periods:
		if bytes > period_bytes:
			period_value, bytes = divmod(bytes, period_bytes)
			# has_s = "s" if period_value > 1 else ""
			# strings.append("%s %s%s" % (period_value, period_name, has_s))
			strings.append(f"{period_value} {period_name}")
			ct += 1
			if max > 0 and ct > max:
				break
	return ", ".join(strings)  # + f"({td_object}, {bytes})"


# Make washee conform strictly to structure of washer
def wash_dict(washee, washer):
	out = {}
	for k, v in washer.items():
		if k in washee:
			if isinstance(v, dict):
				if isinstance(washee[k], dict):
					out[k] = wash_dict(washee[k], v)
			else:
				if not isinstance(washee[k], dict):
					out[k] = washee[k]
	return out


def random_str(l):
	return base64.b64encode(os.urandom(l))


def print_process_info():
	if hasattr(os, "getppid"):
		print(f"Parent process:{os.getppid()}")
	print(f"Process id:{os.getpid()}")


def sleep(time_sec):
	time.sleep(time_sec)
	# await asyncio.sleep async


def read_file(fname, strip=True):
	fn = os.path.join(os.path.dirname(os.path.abspath(__file__)), fname)
	data = ""
	if os.path.exists(fn):
		with open(fn, encoding="utf-8") as f:
			data = f.read()
			data = data.strip() if strip else data
			# logger.info(f"Got data '{data}' from '{fn}'")
	else:
		logger.error(f"Could not find file {fn} relative to working directory {os.getcwd()}")
	return data


def write_file(fn, data, strip=True):
	with open(fn, "w") as f:
		f.write(data.strip() if strip else data)


def debug_imports():
	import logging
	import pprint
	import sys
	import os

	l = logging.getLogger(__name__)
	l.error(f"PATH: {sys.path}")
	l.error(f"PREFIX: {sys.prefix}")
	l.error(f"ENV: {pprint.pformat(os.environ)}")

	import pkg_resources

	installed_packages = {d.project_name: d.version for d in pkg_resources.working_set}
	l.error(f"MODULES:{pprint.pformat(installed_packages)}")

	import pkgutil

	for m in pkgutil.iter_modules(path=None):
		l.error(f"MODULE: {m.name}")

	def explore_package(module_name):
		loader = pkgutil.get_loader(module_name)
		if not loader:
			l.error(f"No loader for {module_name}")
		elif not loader.filename:
			l.error(f"No filename for {module_name}")
		else:
			for sub_module in pkgutil.walk_packages([loader.filename]):
				_, sub_module_name, _ = sub_module
				qname = module_name + "." + sub_module_name
				l.error(qname)
				explore_package(qname)

	explore_package("fk")
	explore_package("batch")

	import time

	time.sleep(die_sec)
	sys.exit(6)


class TimeoutRequestsSession(requests.Session):
	def request(self, *args, **kwargs):
		kwargs.setdefault("timeout", HTTP_TIMEOUT)
		return super(TimeoutRequestsSession, self).request(*args, **kwargs)


def requests_session():
	global _requests_session_instance
	if not _requests_session_instance:
		_requests_session_instance = TimeoutRequestsSession()
		# fmt:off
		headers = {
			  "User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/40.0.2214.85 Safari/537.36"
			, "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9"
			, "Cache-Control": "no-cache"
			, "Pragma": "no-cache"
		}
		# fmt:on
		_requests_session_instance.headers.update(headers)
	return _requests_session_instance


def list_all_pkg():
	for name in sys.modules.keys():
		logger.info(f"Found pacakge: {name}")


def walk_pkg(root: str, base_dir: str = "./", path: str = "", do_debug=False):
	if do_debug and False:
		logger.info(
			f"""walk_pkg(root={root}
, base_dir={base_dir}
, path={path})\n"""
		)
	for name in pkg_resources.resource_listdir(root, path):
		sub_path = f"{path}/{name}"
		full_path = f"{base_dir}{sub_path}"
		if pkg_resources.resource_isdir(root, full_path):
			if do_debug:
				# logger.info(f"{full_path} ------------------\n")
				pass
			walk_pkg(root, base_dir, sub_path, do_debug)
		else:
			# content = pkg_resources.resource_string(root, sub_path)
			if do_debug:
				logger.info(f"  +  {full_path}\n")
			# Do something with the file content


def get_package_file_content(base, filename, do_debug=False):
	"""Look for file relative to package root and return it's content"""
	content = ""
	package_base_path: str = os.path.abspath(os.path.join(os.path.dirname(base), "../"))
	if do_debug:
		logger.warn(f"package_base_path={package_base_path}")
		logger.warn(f"filename={filename}")
		logger.warn(f"__name__={__name__}")
		logger.warn(f"base={base}")
	ex = False
	try:
		ex = pkg_resources.resource_exists(base, filename)
	except Exception as e:
		logger.error(f"Could not find package '{base}'. Se alternatives:")
		list_all_pkg()
	if ex:
		logger.info(f"Fetching {filename} from package")
		content = pkg_resources.resource_string(base, filename).decode("utf-8").strip()
	else:
		logger.warning(f"No content found for {filename}")
	return content


def synchronize(corutine):
	try:
		loop = asyncio.get_running_loop()
	# 'RuntimeError: There is no current event loop...'
	except RuntimeError:
		loop = None
	if loop and loop.is_running():
		tsk = loop.create_task(corutine)
	else:
		asyncio.run(corutine)


def relurl(base):
	out = ""
	if base:
		try:
			base = str(base)
			o = urlparse(base)
			out = o.path
			if o.params:
				out +=o.params
			if o.query:
				out +="?"
				out +=o.query
			if o.fragment:
				out +="#"
				out +=o.fragment
			return out
		except Exception as e:
			logger.exception(e)
			out = ""
	# logger.info(f"'{base}'({type(base)}) --> '{out}'")
	return out


def inject_env(vars):
	# Fake it by inserting values into environment
	for key, val in vars.items():
		os.environ[key] = val
	



def calculate_unique_key(dictionary):
	json_str = json.dumps(dictionary, sort_keys=True)
	unique_key = hashlib.sha256(json_str.encode()).hexdigest()
	return unique_key


def delete_file(filename, do_debug = False):
	if os.path.exists(filename):
		if do_debug:
			logger.info(f"Deleting {filename}")
		os.unlink(filename)
	else:
		if do_debug:
			logger.info(f"Already missing, not deleting {filename}")




def print_caller_info():
	# Get the current stack frame
	stack = inspect.stack()
	
	# The caller is the previous frame in the stack.
	# stack[0] is the current frame, so stack[1] is the caller.
	# You might adjust the index based on your specific needs.
	caller_frame = stack[2]
	
	# Extract information about the caller
	module = inspect.getmodule(caller_frame[0])
	module_name = module.__name__ if module else ''
	filename = caller_frame.filename
	lineno = caller_frame.lineno
	function_name = caller_frame.function
	
	# Print information about the caller
	logger.info(f'INSPECTIFY ### {module_name}::{function_name}() @ {filename}:{lineno}')

def wrap_inspect(method):
	def wrapper(*args, **kwargs):
		print_caller_info()
		return method(*args, **kwargs)
	return wrapper


# Decorate each member function with `wrap_inspect`
def inspectify(clazz):
	for attr_name in dir(clazz):
		attr = getattr(clazz, attr_name)
		if callable(attr) and not attr_name.startswith("__"):
			setattr(clazz, attr_name, wrap_inspect(attr))



import sys
import credentials
from Profiler import Profiler


def file_to_dict(filename):
	ret = {}
	with open(filename, "r") as file:
		line = file.readline()
		while line:
			line = line.strip()
			if line != "":
				ret[line] = line
			line = file.readline()
	print(f"Read {len(ret)} lines from {filename}")
	return ret


# Main entrypoint of script
if __name__ == "__main__":

	with Profiler("Elapsed time {}"):
		i = file_to_dict("i_728k.csv")
		o = file_to_dict("o_440k.csv")
		ret = {}
		for il in i:
			if not bool(o.get(il)):
				ret[il] = il

		out_filename = "out_io.csv"
		with open(out_filename, "w") as outfile:
			for ol in ret:
				bob = f"https://gapi.beeketing.com/v1/recsys/recommendation/result/v2/{ol}/best-seller\n"
				# 				outfile.write(f"{ol}\n");
				outfile.write(bob)
			bob = f"https://gapi.beeketing.com/v1/recsys/recommendation/result/v2/{ol}/best-seller\n"
			print(f"Wrote {len(ret)} lines to {out_filename}")



from enum import Enum
from pydantic import BaseModel, StringConstraints
from typing import List, Optional, Annotated, Any, Dict
import datetime

class JobStatusEnum(str, Enum):
	PENDING = "pending"
	IN_PROGRESS = "in-progress"
	HUNG = "hung"
	DONE = "done"
	FAILED = "failed"

class JobStump(BaseModel):
	'''
	, priority int not null default 50
	, ttl_seconds int
	, data text
	, result text
	, type varchar(255)
	, status varchar(255)
	, throttle_key varchar(63)
	, throttle_limit integer
	, throttle_period integer
	, source text
	, error text
	, worker_id varchar(63) default null
	'''
	priority:int
	ttl_seconds:int | None = None
	data:str | None = None
	result:str | None = None
	type: Annotated[str, StringConstraints(max_length=255)] | None = None
	status: JobStatusEnum | None = None
	throttle_key: Annotated[str, StringConstraints(max_length=63)] | None = None
	throttle_limit:int | None = None
	throttle_period:int | None = None
	source:str | None = None
	error:str | None = None
	runtime:datetime.timedelta | None = None 
	runtime_ts:int | None = None 


class Job(JobStump):
	'''
	  id serial primary key
	, last_started_at timestamptz
	, last_finished_at timestamptz
	, created_at timestamptz not null default now()
	, updated_at timestamptz not null default now()
	'''
	id: int
	last_started_at:datetime.datetime | None = None 
	last_finished_at:datetime.datetime | None = None 
	updated_at:datetime.datetime
	created_at:datetime.datetime


class TypeCount(BaseModel):
	count: int
	type: str

class StatusCount(BaseModel):
	count: int
	status: JobStatusEnum | None = None

class JobCount(BaseModel):
	count: int
	type: str | None = None
	status: JobStatusEnum | None = None

class ItemQuery(BaseModel):
	id: List[int] | None = None
	error: List[str] | None = None
	limit: List[int] | None = None
	priority: List[int] | None = None
	source: List[str] | None = None
	status: List[JobStatusEnum] | None = None
	throttle_key: List[str] | None = None
	ttl_seconds: List[int] | None = None
	type: List[str] | None = None



class JobTimes(BaseModel):
	avg: int | None
	count:int | None
	max:int | None
	max_ttl:int | None
	med:int | None
	min:int | None
	min_ttl:int | None
	type: str | None


class WorkerStats(BaseModel):
	worker_id: str | None
	job_count:int | None
	ms_per_job:int | None
	jobs_per_hour:int | None
	first_active:int | None
	current_job_started:int | None
	last_job_finished:int | None
	run_time:int | None
	work_time:int | None
	idle_time:int | None
	is_activbe:bool | None


class Action(BaseModel):
	action:str
	data:str | None = None



import os
import pprint

from pathlib import Path
import json
import base64

import ast
import octomy.utils
import logging
from functools import lru_cache

from markupsafe import Markup

import yaml

try:
	from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
	from yaml import Loader, Dumper



logger = logging.getLogger(__name__)


class c:
	HEADER = '\033[95m'
	BLUE = '\033[94m'
	CYAN = '\033[96m'
	GREEN = '\033[92m'
	RED = '\u001b[31m'
	WARNING = '\033[93m'
	FAIL = '\033[91m'
	ENDC = '\033[0m'
	BOLD = '\033[1m'
	UNDERLINE = '\033[4m'



# IMPORTANT!!!!!!
# DO NOT PUT REAL CONFIG HERE, ESPECIALLY PASSWORDS AND OTHER SENSITIVE DATA!!!!
unset = "OVERRIDE ME, THIS IS A DEFAULT"

'''

defaults = {
	  "DEBUG": False
	, "TESTING": False
	, "SECRET_KEY": base64.b64encode(os.urandom(16)).decode("utf-8")
	, "SQLALCHEMY_DATABASE_URI": "sqlite:///:memory:"
	, "SQLALCHEMY_ECHO": False
	, "SQLALCHEMY_TRACK_MODIFICATIONS": True
	, "batch-filter-root": "/app/fk/batch/filters"
	, "cache-dir-name": "_cache"
	, "crawler-html-parser": "html5lib"
	, "crawler-referer": "https://www.shopify.com/"
	, "crawler-socket-timeout": 15
	, "crawler-user-agent": "Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)"
	, "db-database": "postgres"
	, "db-hostname": "localhost"
	, "db-password": unset
	, "db-port": "5432"
	, "db-username": unset
	, "db-sql-dir": "sql"
	, "smtp-hostname": unset
	, "smtp-port": "2525"
	, "smtp-user": unset
	, "smtp-password": unset
	, "smtp-headers": unset
	, "influx-dbname": "sales"
	, "influx-hostname": "localhost"
	, "influx-password": unset
	, "influx-port": "8086"
	, "influx-username": unset
	, "n2report-dir-name": "_n2report"
	, "n2report-index-name": "index.html"
	, "overwatch-password": unset
	, "overwatch-user": unset
	, "preferred-url-scheme": "http"
	, "printful-api-endpoint": "https://api.printful.com/"
	, "printful-api-key": unset
	, "printful-api-key-base64": unset
	, "screenshots-dir-name": "_screenshots"
	, "screenshots-index-name": "screenshots.html"
	, "shop-id-generator-batch-size": 200
	, "shop-id-generator-max": 25000000000  # billions
	, "shop-id-generator-min": 0
	, "shop-id-thread-count": 300
	, "shopify-api-key": unset
	, "shopify-api-secret": unset
	, "shopify-api-throttle-limit": 1
	, "shopify-api-throttle-period": 3000
	, "shopify-api-version": "2020-10"
	, "shopify-app-domain": unset
	, "shopify-app-author": "The author of this app"
	, "shopify-app-author-email": "The email of the author of this app"
	, "shopify-app-company-name": "The name of the company behind this app"
	, "shopify-app-company-website": "The url fo the company website"
	, "shopify-app-description": "The description of this app"
	, "shopify-app-name": "The name of this app"
	, "shopify-app-preferred-url-scheme": "https"
	, "shopify-password": unset
	, "shopify-shared-secret": unset
	, "static-files-root": "/app/shopify_app/static"
	, "webdriver-max-windows": 5
	, "webdriver-url": "http://127.0.0.1:4444/wd/hub"
	, "webdriver-wait-time-sec": 10
	
	
	, "app-name": Markup("Korall&trade;")
	, "app-url": Markup("https://korall.io")
	, "app-contact-email": Markup("contact@korall.io")
	, "app-contact-telephone": Markup("+47 942 37 478")
	, "app-webmaster-email": Markup("webmaster@korall.io")
	, "app-linkedin": Markup("korall-engineering-as")
	, "app-instagram": Markup("korall_engineering")
	, "app-logo": "/static/images/Korall_Engineering_logo.svg"
	, "app-logo-dark": "/static/images/Korall_Engineering_logo_dark.svg"
	, "app-slogan": Markup("Challenge creation&trade;")
	, "app-geo-lat": 60.37776008039476
	, "app-geo-lon": 5.331693878840272
	, "app-geo-zoom": 19
	, "app-description": Markup("Korall Engineering AS is a design engineering firm, part of StartupLab Bergen, that employs the latest high-end product optimization technologies")
	, "app-admin-path": "overwatch"

	, "app-unknown-user-image": "/static/images/unknown-user.svg"
	, "app-db-fetch-limit": 1000
	
	# "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/40.0.2214.85 Safari/537.36"
	# SHOPIFY_SHOP_NAME = 'firstkissdesigns'
	# https://flask.palletsprojects.com/en/1.1.x/config/#DEBUG
	# https://flask.palletsprojects.com/en/1.1.x/config/#SECRET_KEY
	# SERVER_NAME = "localhost.localdomain:5000"
	# https://flask.palletsprojects.com/en/1.1.x/config/#TESTING
}

'''


class OctomyConfig(dict):
	def __init__(self):
		super().__init__()
		self.sources = dict()


	def TO_BOOL(str):
		"""Helper to parse boolean from environment variable"""
		if str.lower() in {"1", "t", "true"}:
			return True
		elif str.lower() in {"0", "f", "false"}:
			return False
		return None

	# With help from https://stackoverflow.com/questions/1265665/how-can-i-check-if-a-string-represents-an-int-without-using-try-except
	def TO_UINT(str):
		"""Helper to parse unsigned integer from environment variable"""
		if str.isdigit():
			return int(str)
		if str.startswith("+") and str[1:].isdigit():
			return int(str)
		return None


	def TO_INT(str):
		"""Helper to parse signed integer from environment variable"""
		if str.isdigit():
			return int(str)
		if str.startswith(("+", "-")) and str[1:].isdigit():
			return int(str)
		return None


	def TO_ENV_KEY(str):
		"""Helper to convert a string to environment variable name on the form THIS_IS_ENV_VARIABLE"""
		str = str.strip().upper()
		for i in ["-", " ", "\t"]:
			str = str.replace(i, "_")
		return str


	def TO_GUESS_TYPE(str):
		"""Helper to convert a string to a native python type"""
		b = OctomyConfig.TO_BOOL(str)
		if b is not None:
			return b
		ret = str
		try:
			ret = ast.literal_eval(str)
		except:
			pass
		return ret


	def apply_environment_variables(self, do_debug=False):
		"""Helper to load configuration values from environment variables"""
		if do_debug:
			logger.info(f"apply_environment_variables()")
		ct = 0
		for key, value in self.items():
			# Try with bare key first
			env_value = os.environ.get(key, default=None)
			if do_debug:
				logger.info(f" {key}: {env_value}")
			if not env_value:
				# Try with "envified" key second
				env_key = OctomyConfig.TO_ENV_KEY(key)
				env_value = os.environ.get(env_key, default=None)
				if do_debug:
					logger.info(f" {env_key}: {env_value}")
			if env_value:
				old_value = self.get(key)
				old_source = self.sources.get(key)
				if old_value:
					if do_debug:
						logger.info(f"Overwriting existing value {old_value} from {old_source}")
				self[key] = OctomyConfig.TO_GUESS_TYPE(env_value)
				self.sources[key] = "environment"
				ct += 1
				# logger.info(f"Value from env '{env_value}' replaced '{value}' for key '{key}'")
			else:
				# logger.warning(f"WARNING: No {env_key} in env, using last value '{key}' = {value}")
				pass
		logger.info(f"Put {ct} values from environment variables into configuration")
		return True

	def apply_dict(self, new_values, source="dict", do_debug=False):
		"""Helper to load configuration values from dictionary"""
		if do_debug:
			logger.info(f"apply_dict(other=, source={source})")
		ct = 0
		if not new_values:
			logger.warning(f"No values for '{source}'")
			return ct
		for key, value in new_values.items():
			old_value = self.get(key)
			old_source = self.sources.get(key)
			if do_debug:
				logger.info(f" {key}:")
			if old_value:
				if do_debug:
					logger.info(f" <-- {old_value} ({old_source})")
			if do_debug:
				logger.info(f" --> {value} ({source})")
			self[key] = value
			self.sources[key] = source
			ct += 1
			# logger.info(f"Value from file '{file_value}' replaced '{value}' for key '{key}'")
		logger.info(f" Put {ct} values from '{source}' into configuration")
		return ct

	def apply_json_config_file(self, filename, do_debug=False):
		"""Helper to load configuration values from configuration file in json format"""
		if do_debug:
			logger.info(f"apply_json_config_file(filename={filename})")
		file = Path(filename)
		# Create a config file with defaults if none exists
		if not file.is_file():
			logger.info(f" Config file '{filename}' did not exist, skipping")
			logger.info(f" NOTE: current dir was: {os.getcwd()}")
			return False
		# Load config from file
		loaded_config = {}
		if not os.path.exists(filename):
			logger.error(f"ERROR: Could not find '{filename}'")
			return False
		with open(filename, "r") as json_file:
			logger.info(f" Loading config from '{filename}'")
			try:
				loaded_config = json.load(json_file)
				# logger.info(f"LOADED JSON: {pprint.pformat(loaded_config)}")
				self.apply_dict(loaded_config, filename, do_debug)
			except Exception as e:
				logger.error(f"\nParsing json from config file '{filename}' failed:\n{e}\n", exc_info=True)
				return False
		return True


	def apply_yaml_config_file(self, filename, do_debug=False):
		"""Helper to load configuration values from configuration file in yaml format"""
		if do_debug:
			logger.info(f"apply_yaml_config_file(filename={filename})")
		file = Path(filename)
		# Create a config file with defaults if none exists
		if not file.is_file():
			logger.info(f" Config file '{filename}' did not exist, skipping")
			logger.info(f" NOTE: current dir was: {os.getcwd()}")
			return False
		# Load config from file
		loaded_config = {}
		if not os.path.exists(filename):
			logger.error(f"ERROR: Could not find '{filename}'")
			return False
		with open(filename, "r") as json_file:
			logger.info(f" Loading config from '{filename}'")
			try:
				loaded_config = yaml.safe_load(json_file)
				# logger.info(f"LOADED YAML: {pprint.pformat(loaded_config)}")
				self.apply_dict(loaded_config, filename, do_debug)
			except Exception as e:
				logger.error(f"\nParsing yaml from config file '{filename}' failed:\n{e}\n", exc_info=True)
				return False
		return True


	def list_unchanged(self):
		""" List all variables that have not been changed from their default setting"""
		logger.info(f"Unchanged configurations options:")
		for key, value in defaults.items():
			if self.get(key) == value:
				logger.info(f"{key}={value}")

	def members(self):
		""" Return all members of the config"""
		ret = {}
		base = self  # .__class__
		for member in [attr for attr in dir(base) if not callable(getattr(base, attr)) and not attr.startswith("__")]:
			val = getattr(base, member)
			ret[member] = val
		return ret

	def redact_key(self, key, val, num=20, source="redact"):
		"""Redact the value of a key for security"""
		if source in ["default", "none"]:
			return val
		if any(x in key.lower() for x in ["secret", "pass", "pwd", "key", "psk"]):
			val = octomy.utils.redact(val, num)
		return val

	def __str__(self):
		"""Print intuitive overview of settings"""
		dict = self.members()
		ret = f"\n########## Configuration options for {self.__class__.__name__}:\n"
		order = {"none": {}, "default": {}}
		for key, val in dict.items():
			if self.sources != val:
				source = self.sources.get(key, "none")
				order[source] = order.get(source, {})
				order[source][key] = val
		for source, items in order.items():
			if len(items) > 0:
				ret += f"[{c.BLUE}{source}{c.ENDC}]\n"
				for key, val in items.items():
					vc = c.GREEN
					rv = self.redact_key(key, val)
					if val == unset:
						vc = c.RED
					if rv != val:
						vc += c.BOLD
					ret += f" {c.BLUE}+ {c.CYAN}{key} {c.ENDC}= {vc}{rv}{c.ENDC}\n"
		return ret

	def redacted_dict(self) -> dict:
		"""Return the full config with all values redacted for security"""
		out_dict = dict(self.members())
		for key, val in out_dict.items():
			out_dict[key] = self.redact_key(key, val)
		return out_dict

	def attrify(self):
		""" Convert all config values in the dictionart into attributes"""
		for key, val in self.items():
			# logger.debug(f" + {key} = {self.redact_key(key, val)}")
			setattr(self.__class__, key, val)



class DefaultOctomyConfig(OctomyConfig):

	def __init__(self):
		"""The default base configuration to be extended by actual configurations to be used"""
		super().__init__()
		self.update(defaults)
		self.sources = {k: "default" for k in self.keys()}


	def verify(self):
		"""Helper to check validity of configuration"""
		defkey = []
		for key in ["printful-api-key"]:
			if self.get(key, None) == defaults.get(key, None):
				defkey.append(key)
		if defkey:
			logger.warning(f"Config contains default values for the following keys: '{', '.join(defkey)}'\n        If there are any authentication error this might be the culprit.")

		# Printify needs printful-api-key base64 encoded, so do that now b64encode
		pak = self.get("printful-api-key", None)
		if pak:
			self["printful-api-key-base64"] = str(base64.urlsafe_b64encode(pak.encode("utf-8")).strip(), "utf-8")


@lru_cache()
def get_config() -> OctomyConfig:
	config = OctomyConfig()
	do_debug = False
	config.apply_yaml_config_file("config/organization.yaml", do_debug)
	config.apply_yaml_config_file("config/application.yaml", do_debug)
	config.apply_yaml_config_file("config/service.yaml", do_debug)
	config.apply_yaml_config_file("config/local_secrets.yaml", do_debug)
	config.apply_environment_variables(do_debug)
	#config.verify()
	config.attrify()
	logger.info(config)
	return config



from fastapi import Depends
from functools import lru_cache
from pydantic_settings import BaseSettings
from typing import Any, Callable, Set, List, Optional, Annotated
import logging
import octomy.access
import octomy.config
import octomy.db
import octomy.batch
import pprint
import asyncio

logger = logging.getLogger(__name__)


DEFAULT_ADMIN_PATH="overwatch"

def url_for(name:str) -> str:
	url = f"http://CONTEXT_URL_FOR/{name}"
	logger.info(f"Generated url '{url}'")
	return  url



#required_scopes=["overwatch.access.profile.view"]
#required_scopes: List[str]

class Context(BaseSettings):
	config: octomy.config.OctomyConfig | None
	access: octomy.access.AccessContext | None
	db: octomy.db.Database | None
	bp: octomy.batch.Processor | None
#	current_user: Optional[octomy.access.User] = octomy.access.get_current_user()
#	url_for:Callable[[str], str] = url_for # SUBSTITUTE WITH NATIVE FASTAPI request.url_for() INSTEAD
	admin_path:str = DEFAULT_ADMIN_PATH
	required_scopes:List[str] = list()
	#@property
	#def url_for(self, key) -> str:
	#	return url_for(key)
#	data:Data = Data()
#	parts:Parts = Parts()
#	search_engine:SearchEngine = SearchEngine()
	webroot: str = "/app/webroot"
	password_cost_factor:int = 12


def default_decorator_function(undecorated_context:Context):
	return undecorated_context

_dec_fun = default_decorator_function

def set_context_decorator(dec_fun):
	_dec_fun = dec_fun



def scoped_context(s:list):
	return Annotated[Context, Depends(context_dependency(required_scopes=s))]



#@lru_cache
def get_context(required_scopes: Optional[List[str]] = None) -> Context:
	config:octomy.config.OctomyConfig = octomy.config.get_config()
	db:octomy.db.Database = octomy.db.get_database(config)
	bp:octomy.batch.Processor = octomy.batch.Processor(config, db)
	access:octomy.access.AccessContext = octomy.access.get_access_context(required_scopes)
	logger.info(f"AccessContext: {pprint.pformat(access)}")
	return Context(config=config, access=access, db=db, bp=bp, admin_path = config.get("app-admin-path", DEFAULT_ADMIN_PATH))


# Helper dependency
def context_dependency(required_scopes: Optional[List[str]] = None):
	async def dependency():
		return get_context(required_scopes=required_scopes)
	return dependency




from fastapi import Request, HTTPException
from fastapi.responses import HTMLResponse
from pathlib import Path
import datetime
#import octomy.utils
#import octomy.utils.expiry_cache
import jinja2
import logging
import os
import pprint
import traceback


logger = logging.getLogger(__name__)

def gen_tags(parts):
	tags = ["-".join(parts[:-1])] + [part for part in parts]
	# logger.info(f"TAGS: {pprint.pformat(parts)} --> {tags}")
	return  tags




def url_for(index):
	return "http://TODO_MAKE_URLFOR/"+str(index)

class AutoRouter:

	def __init__(self, app, context):
		self.app = app
		self.context = context
		self.env = self.prepare_env()
		self.auto_routes = {}
		self.auto_route_all(self.context.webroot)
		#logger.info(f"ROUTES: {pprint.PrettyPrinter(indent=4, compact=False).pformat(self.auto_routes)}")


	def prepare_env(self):
		loader = jinja2.ChoiceLoader(
			[
				jinja2.FileSystemLoader(self.context.webroot)
	#			, jinja2.FileSystemLoader(os.path.join(project_folder, os.path.dirname(template_name)))
			]
		)
		env = jinja2.Environment(loader=loader)
		env.globals['datetime'] = datetime
		env.globals['context'] = self.context
#		env.globals['current_user'] = user
#		env.globals['current_user_can'] = current_user_can
#		env.globals['url_for'] = url_for
		#env.globals.update(get_expiry=octomy.utils.expiry_cache.get_expiry)
		#env.filters['myfilter'] = filter
		return env


	def subpages(self, path, max_level=2):
		out = {}
		#space = max_level * "   "
		parts = os.path.normpath(path).strip(os.path.sep).split(os.path.sep) 
		path = "/".join(parts)
		obj = self.auto_routes.get(path, dict())
		if max_level > 1:
			out[path] = obj
			if obj:
				subs = obj.get("children", list())
				for sub in subs:
					subsubs = self.subpages(sub, max_level - 1)
					for subsub in subsubs:
						if subsub not in out:
							subobj = self.auto_routes.get(subsub, dict())
							if subobj:
								out[subsub] = subobj
		return out


	def handler_raw(self, path):
		try:
			template_fn = f"{self.context.webroot}/{path}/index.html"
			template_string = ""
			parts = os.path.normpath(path).strip(os.path.sep).split(os.path.sep) 
			#dir = Path(os.path.relpath(os.path.dirname(template), self.context.webroot))
			subs = self.subpages(path, 4)
			#return f"subpages {pprint.pformat(subs)}", None
			logger.info(f"Reading template {template_fn}:")
			with open(template_fn, "r") as template_file:
				template_string = template_file.read()
				#logger.info(f"Content was: {template_string}")
			#return template_string, None
			template = self.env.from_string(template_string)
			if not template:
				return None, f"Error creating template for '{template_fn}'"
			#logger.info(f"template_fn:{template_fn}")
			html = template.render(title="TODO: Set title from context", subpages=subs)
			return html, None
		except Exception as e:
			logger.exception(f"Could not handle {path}")
			return None, f"Error rendering template '{template_fn}' with error\n{traceback.format_exc()}"


	def auto_route_all(self, base):
		for path in Path(base).rglob('index.html'):
			self.auto_route_path(path)
	

	def auto_route_path(self, template_fn):
		dir = Path(os.path.relpath(os.path.dirname(template_fn), self.context.webroot))
		self.auto_routes[str(dir)] = { "children":[], "path": str(dir), "title": os.path.basename(str(dir)).capitalize() }
		name = os.path.basename(dir)
		#logger.info(f"GOT ROUTE{pprint.pformat(dir.parts)} {name}")
		self.app.add_api_route(path=f"/{dir}", endpoint=self.handler, methods=["GET"], tags=gen_tags(dir.parts), name=name, response_class=HTMLResponse)
		acc = []
		for part in dir.parts:
			key = "/".join(acc)
			node = self.auto_routes.get(key, dict())
			children = node.get("children", list())
			childkey = f"{key}/{part}"
			if childkey not in children:
				children.append(childkey)
			node["children"] = children
			node["path"] = str(key)
			node["title"] = os.path.basename(str(key)).capitalize()
			self.auto_routes[key] = node
			#index_page(str(key))
			acc.append(part)
		return {"dir":dir, "name":name, }
	

	def handler(self, request:Request):
		path = request.scope['route'].path
		logger.info(f"GOT HANDLER{pprint.pformat(path)}, {pprint.pformat(request.path_params)} {request.url.path}")
		res, err =  self.handler_raw(path)
		if err:
			return HTTPException(status_code=500, detail=err)
		return HTMLResponse(content=res, status_code=200)



from fk.batch.BatchProcessor import BatchProcessor
from multiprocessing import Process, connection
import os
import signal

import logging

logger = logging.getLogger(__name__)


class Server:
    def __init__(self, config):
        self.config = config
        self.term_received = False

        def handler(signum, frame):
            logger.info(f"{signum} signal received, terminating gracefully")
            self.term_received = True

        signal.signal(signal.SIGABRT, handler)

    def verify(self):
        batch_processor = BatchProcessor(config=self.config)
        return batch_processor.verify()

    def wrapper(self, num):
        logger.info(f"Batch worker {num} started with id:{os.getpid()}, parent:{os.getppid()}")
        batch_processor = BatchProcessor(config=self.config)
        while not self.term_received:
            # Do some work
            batch_processor.process()
        logger.info(f"Worker {num} stopped")

    def start(self, num):
        logger.info(f"Batch worker {num} starting...")
        worker = Process(target=self.wrapper, args=(num,))
        self.workers.append(worker)
        worker.start()
        return worker

    # Start server and serve forever
    def run(self):
        self.workers = []
        workers_count = self.config.get("batch-workers", 1)
        logger.info(f"Starting {workers_count} workers")
        num = 0
        for _ in range(workers_count):
            self.start(num)
            num += 1
        # Restart workes that terminate for whatever reason (they will self terminate on errors)
        while not self.term_received:
            connection.wait(worker.sentinel for worker in self.workers)
            for worker in self.workers:
                logger.info(f"Batch worker terminating...")
                worker.join()
                self.workers.remove(worker)
                logger.info(f"Batch worker terminated")
                self.start(num)
                num += 1
        logger.info(f"Server stopping:")
        for worker in self.workers:
            logger.info(f"Batch worker terminating...")
            worker.join()
            self.workers.remove(worker)
        logger.info(f"Server stopped")



from pydantic import BaseModel, StringConstraints
from typing import List, Optional, Annotated, Any, Dict
import pydantic.version
import pytz
from .types import JobCount, StatusCount, TypeCount, Job, JobTimes, WorkerStats
import logging
import pprint


logger = logging.getLogger(__name__)

class Database:



	def __init__(self, config, dbc):
		self.config = config
		self.dbc = dbc


	# Look at faster batch imports with
	# from psycopg2.extras import execute_values
	# execute_values(cur,
	# "INSERT INTO test (id, v1, v2) VALUES %s",
	# [(1, 2, 3), (4, 5, 6), (7, 8, 9)])
	# FROM https://stackoverflow.com/questions/8134602/psycopg2-insert-multiple-rows-with-one-query


	# Insert a new batch item into log
	#async def insert_batch_item(self, batch_item):
	#	return await self.dbc.key_query(query_key = "overwatch.batch.insert_batch_item", params=batch_item, mode = "one", item_type = dict)


	# Update status of all "in progress" jobs that already spent more than ttl time to "hung"
	#async def update_batch_hang_status(self):
	#	updated_status= {"from_status": (self.IN_PROGRESS,), "to_status": (self.HUNG,), "in_progress_status": (self.IN_PROGRESS,)}
	#	return await self.dbc.key_query(query_key = "overwatch.batch.update_hang_status", params=updated_status, mode = "one", item_type = dict)

	# Update status of all jobs with from_status to to_status
	#async def bump_batch_items(self, from_status=IN_PROGRESS, to_status=PENDING):
	#	updated_status = {"from_status": (from_status,), "to_status": (to_status,), "in_progress_status": (self.IN_PROGRESS,)}
	#	return await self.dbc.key_query(query_key = "overwatch.batch.bump_batch_items", params=updated_status, mode = "one", item_type = dict)

	# Update status of a batch item in the batch log, making sure to fail if the id and updated_at don't match, providing a guarantee of atomic operation
	# Returns updated_at, so caller can check if it was updated or not (compare it to argument)
	async def bump_batch_item(self, id, status, error, updated_at, result=None):
		updated_status = {"id": (id,), "status": (status,), "error": (error,), "result": (result,), "updated_at": (updated_at,), "in_progress_status": (self.IN_PROGRESS,)}
		return await self.dbc.key_query(query_key = "overwatch.batch.bump_batch_item", params=updated_status, mode = "one", item_type = dict)

	# Simplified version of bump_batch_item where status is updated, no questions asked
	async def bump_batch_item_status(self, id, status):
		updated_status = {"id": (id,), "status": (status,), "in_progress_status": (self.IN_PROGRESS,)}
		return await self.dbc.key_query(query_key = "overwatch.batch.bump_batch_item_status", params=updated_status, mode = "one", item_type = dict)

	# Update a batch item in the batch log, making sure to fail if the id and updated_at don't match, providing a guarantee of atomic operation
	#async def book_batch_item(self, from_status=PENDING, to_status=IN_PROGRESS, worker_id=None):
	#	updated_status = {"from_status": (from_status,), "to_status": (to_status,), "in_progress_status": (self.IN_PROGRESS,), "worker_id": (worker_id,)}
	#	return await self.dbc.key_query(query_key = "overwatch.batch.book_batch_item", params=updated_status, mode = "one", item_type = dict)

	# Update a batch item in the batch log, making sure to fail if the id and updated_at don't match, providing a guarantee of atomic operation
	# NOTE: This is the throttled version
	#async def book_throttled_batch_item(self, from_status=PENDING, to_status=IN_PROGRESS, worker_id=None):
	#	updated_status = {"from_status": (from_status,), "to_status": (to_status,), "in_progress_status": (self.IN_PROGRESS,), "worker_id": (worker_id,)}
	#	return await self.dbc.key_query(query_key = "overwatch.batch.book_throttled_batch_item", params=updated_status, mode = "one", item_type = dict)

	# Delete single item by id
	#async def delete_by_id(self, id):
	#	query = {"id": (id,)}
	#	db_items, db_err = await self.dbc.key_query(query_key = "overwatch.batch.delete_by_id", params = query, mode = "none", do_debug=True)
	#	return {}

	# Delete batch items with given status and update_at longer than given time
	#async def delete_batch_items_with_status(self, status):
	#	query = {"status": (status,)}
	#	db_items, db_err = await self.dbc.key_query(query_key = "overwatch.batch.delete_batch_items_by_status", params = query, mode = "none", do_debug=True)
	#	return {}

	# Clear out the batch items table
	#async def delete_all(self):
	#	query = {}
	#	db_items, db_err = await self.dbc.key_query(query_key = "overwatch.batch.delete_all", params = query, mode = "none", do_debug=True)
	#	return {}

	# Get distinct batch types with counts
	async def get_type_counts(self) -> List[TypeCount]:
		query = {}
		return await self.dbc.key_query(query_key="overwatch.batch.get_type_counts", params=query, mode = "many", item_type = List[TypeCount])

	# Get distinct batch status with counts
	#async def get_status_counts(self) -> List[StatusCount]:
	#	query = {}
	#	return await self.dbc.key_query(query_key="overwatch.batch.get_status_counts", params=query, mode = "many", item_type = List[StatusCount])

	# Get distinct batch status+type with counts
	#async def get_job_counts(self) -> List[JobCount]:
	#	query = {}
	#	return await self.dbc.key_query(query_key="overwatch.batch.get_job_counts", params=query, mode = "many", item_type = List[JobCount])

	# Get statistics for each worker (inferred)
	async def get_worker_stats(self, limit=10) -> List[WorkerStats]:
		query = {"limit": limit}
		return await self.dbc.key_query(query_key="overwatch.batch.get_worker_stats", params=query, mode = "many", item_type = List[WorkerStats])

	# Get progression stats with totals for the given statuses
	#async def get_job_times(self, statuses=["done"]) -> List[JobTimes]:
	#	query = {"statuses": statuses}
	#	return await self.dbc.key_query(query_key="overwatch.batch.get_job_times", params=query, mode = "many", item_type = List[JobTimes])

	# Get batch items from batch log sorted by last active
	#async def get_jobs(self, id=None, priority=None, ttl_seconds=None, type="tits", status=None, throttle_key=None, source=None, error=None, limit=1):
	#	query = {"id": id, "priority": priority, "ttl_seconds": ttl_seconds, "type": type, "status": status, "throttle_key": throttle_key, "error": error, "source": source, "limit": limit}
	#	logger.info(f"DB PARAMS: {pprint.pformat(query)}")
	#	return await self.dbc.key_query(query_key="overwatch.batch.get_jobs", params=query, mode = "many", item_type = List[Job])


	#    , min( extract('epoch' from updated_at - created_at)) as low
	# , max( extract('epoch' from updated_at - created_at)) as high
	# , percentile_cont(0.5) within group (order by extract('epoch' from "test_median".updated_at - "test_median".created_at)::float ) as median

	# Return database's idea of current time
	async def get_now(self):
		query = {}
		db_items, db_err = await self.dbc.key_query(query_key = "overwatch.batch.get_now", params=query, mode = "one", item_type = dict)
		db_items = db_items.replace(tzinfo=pytz.UTC)
		return db_items



#!/usr/bin/env python

import os
import pprint
from setuptools import setup, find_namespace_packages

import logging

logger = logging.getLogger(__name__)

setup_requirements = ["pytest-runner", "setuptools_scm"]

release_branch = "development"
beta_branch = "beta"

version_file = "./VERSION"
license_file = "./LICENSE"
changelog_file = "./CHANGELOG"
readme_file = "./README.md"
code_dir="./"
python_version = "3.9"

group_base_name = "octomy"
base_name = "common"
modules = ["octomy"] # group_base_name
short_description = (group_base_name+"/"+base_name),
package_name = f"{group_base_name}-{base_name}"

author_name = "OctoMY"
author_email = "pypi@octomy.org"


def read_file(fname, strip=True):
	fn = os.path.realpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), fname))
	data = ""
	if os.path.exists(fn):
		with open(fn) as f:
			data = f.read()
			data = data.strip() if strip else data
			# logger.info(f"Got data '{data}' from '{fn}'")
	else:
		logger.error(f"Could not find file {fn}")
		logger.warning(f"NOTE: Current working directory is {os.getcwd()}")
	return data


def write_file(fname, data, do_overwrite=False):
	fn = os.path.realpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), fname))
	if not os.path.exists(fn) or do_overwrite:
		with open(fn, "w") as f:
			f.write(data)
	else:
		logger.warning(f"File {fn} already exists")
		logger.warning(f"NOTE: Current working directory is {os.getcwd()}")
	return data


def remove_comment(line, sep="#"):
	i = line.find(sep)
	if i >= 0:
		line = line[:i]
	return line.strip()


def read_requirements_file(fname: str, do_strip: bool = True):
	fn = os.path.realpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), fname))
	print(f"Reading requirements from {fn} with do_strip = {do_strip}")
	lines = []
	if os.path.exists(fn):
		with open(fn) as f:
			for r in f.readlines():
				r = r.strip()
				if len(r) < 1:
					continue
				r = remove_comment(r)
				if len(r) < 1:
					continue
				lines.append(r)
	else:
		logger.error(f"Could not find requirements file {fn}")
		logger.warning(f"NOTE: Current working directory is {os.getcwd()}")
	# logger.warning(f"Full content of '{fname}' was: \n{lines}")
	if not do_strip:
		return lines
	out = []
	for line in lines:
		if line and not line.startswith("-"):
			out.append(line)
	return out


def debug_repo(repo):
	if not repo:
		print(f"No repo")
		return
	print(f"Repository head commit: {repo.head.commit}")
	print(f"Found {len(repo.branches)} branches:")
	for branch in repo.branches:
		print(f" + {branch}({branch.commit})")
	remote = repo.remote()
	print(f"Found {len(remote.refs)} remote refs:")
	for ref in remote.refs:
		print(f" + {ref}({ref.commit})")


def get_git_branch_from_env():
	branch_env = "FK_GIT_ACTUAL_BRANCH"
	branch = os.environ.get(branch_env, None)
	if branch is not None:
		print(f"Using {branch_env} = {branch} from environment")
	else:
		print(f"No value for {branch_env} found")
	return branch


def get_license_name():
	fn = os.path.realpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), license_file))
	print(f"Reading license from {fn}")
	if os.path.exists(fn):
		with open(fn) as f:
			for r in f.readlines():
				r = r.strip()
				if len(r) < 1:
					continue
				# Return first non-empty line
				return r
	else:
		logger.error(f"Could not find license file {fn}")
		logger.warning(f"NOTE: Current working directory is {os.getcwd()}")
	return "Proprietary" # Fall back to something safe


def generate_version_string(version=None, branch=None):
	version = read_file(version_file) if version is None else version
	branch = get_git_branch_from_env() if branch is None else branch
	full_version = ""
	# Calculate the version string based on current branch and bare version
	# TODO: Not tested
	if branch == release_branch:
		full_version = version
	elif branch == beta_branch:
		full_version = f"{version}b"
	elif "feature-" in branch:
		feature_branch_name = branch.replace("feature-", "")
		full_version = f"{version}-{feature_branch_name}"
	else:
		full_version = f"{version}-nonprod"
	print(f"Using full version = {full_version}")
	return full_version

def generate_development_status(version=None, branch=None):
	version = read_file(version_file) if version is None else version
	branch = get_git_branch_from_env() if branch is None else branch
	development_status = ""
	# Calculate the version string based on current branch and bare version
	if branch == release_branch:
		development_status = "Development Status :: 5 - Production/Stable"
	elif branch == beta_branch:
		development_status = "Development Status :: 4 - Beta"
	elif "feature-" in branch:
		development_status = "Development Status :: 3 - Alpha"
	else:
		development_status = "Development Status :: 1 - Planning"
	return development_status

def get_version_string():
	# Not viable
	# return generate_version_string();
	return read_file(version_file)


def get_development_status():
	# Not viable
	# return generate_development_status();
	return "Development Status :: 1 - Planning"



def get_license():
	bob={
		"license": "Proprietary",
	}



package = {
	  "name": package_name
	, "version": get_version_string()
	, "author": author_name
	, "author_email": author_email
	, "maintainer": author_name
	, "maintainer_email": author_email
	, "description": short_description
	, "license": get_license_name()
	, "platforms": ["Linux"]
	, "keywords": "software"
	, "url": f"https://gitlab.com/{group_base_name}/{base_name}"
	, "packages": find_namespace_packages(where=code_dir, include=[module+".*" for module in modules])
	, "package_dir": {'': code_dir}
	, "namespace_packages": modules
	, "long_description": read_file(readme_file)
	, "long_description_content_type": "text/markdown"
	, "setup_requires": setup_requirements
	, "zip_safe": True
	# Allow flexible deps for install
	, "install_requires": read_requirements_file("requirements/requirements.in")
	# Use flexible deps for testing
	, "tests_require": read_requirements_file("requirements/test_requirements.in")
	, "test_suite": os.path.join(code_dir, "tests")
	, "python_requires": ">=" + python_version
	# Needed because installed package depends on this file to know it's version number
	, "data_files": [(base_name, [version_file]), (base_name, [license_file])]
	, "include_package_data": True
	# From https://pypi.org/pypi?%3Aaction=list_classifiers
	, "classifiers": [
		  get_development_status()
		, "Intended Audience :: Developers"
		, "Intended Audience :: Information Technology"
		, "Intended Audience :: Science/Research"
		, "Intended Audience :: Other Audience"
		, "Topic :: Utilities"
		, "Natural Language :: English"
		, "Operating System :: POSIX :: Linux"
		, "Programming Language :: Python :: " + python_version
		, "Topic :: Other/Nonlisted Topic"
	]
}

print("-------------------------------------------------------")
print("setup.py package:")
print(pprint.pformat(package))
print("-------------------------------------------------------")

# pprint.pprint(package)
setup(**package)



from pathlib import Path
import json
import sys
import os
import base64
from .utils import merge

# IMPORTANT!!!!!!
# DO NOT PUT REAL CREDENTIALS HERE!
# INSTEAD EDIT credentials.json THAT IS
# AUTOMATICALLY CREATED ON FIRST RUN OF MAIN SCRIPT
default_credentials = {
	"printful-api-key-source": "Get it here: https://firstkissdesign.myshopify.com/admin/apps/3dd87416a3b1e12a9ef3604aaaaf4d1e/dashboard/settings/1526227/store-api",
	"printful-api-key": "REPLACE-WITH-ACTUAL-KEY",
	"printful-api-key-base64": "",
	"printful-api-endpoint": "https://api.printful.com/",
	"shopify-api-key": "",
	"shopify-password": "",
	"shopify-api-version": "",
	"shopify-shared-secret": "",
	"shopify-shop-name": "",
	"webdriver-url": "http://127.0.0.1:4444/wd/hub",
	"webdriver-max-windows": 5,
	"webdriver-wait-time-sec": 10,
	"influx-hostname": "localhost",
	"influx-port": "8086",
	"influx-username": "root",
	"influx-password": "root",
	"influx-dbname": "sales",
	"cache-dir-name": "_cache",
	"n2report-dir-name": "_n2report",
	"n2report-index-name": "index.html",
	"screenshots-dir-name": "_screenshots",
	"screenshots-index-name": "screenshots.html",
	"db-hostname": "localhost",
	"db-port": "5432",
	"db-username": "postgres",
	"db-password": "",
	"db-database": "postgres",
	"shop-id-generator-min": 0,
	"shop-id-generator-max": 25000000000,  # billions
	"shop-id-generator-batch-size": 200,
	"shop-id-thread-count": 300,
	"crawler-socket-timeout": 15,
	"crawler-referer": "https://www.shopify.com/",
	"crawler-user-agent": "Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)",
	# "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/40.0.2214.85 Safari/537.36"
	"crawler-html-parser": "html5lib",
}


# Load credentials from file, making sure to create defaults if no credentials file found
def load_credentials_ba(credentials_fn="credentials.json"):
	credentials_file = Path(credentials_fn)
	# Create a credentials file with defaults if none exists
	if not credentials_file.is_file():
		print(f"Credentials file {credentials_fn} did not exist, creating")
		with open(credentials_fn, "w") as outfile:
			json.dump(default_credentials, outfile, ensure_ascii=False, sort_keys=True, indent=3)
	# Load credentials from file
	credentials = {}
	if not os.path.exists(credentials_fn):
		print(f"ERROR: Could not find {credentials_fn}")
		sys.exit(1)
	with open(credentials_fn, "r") as json_file:
		print(f"Loading credentials from {credentials_fn}")
		try:
			loaded_credentials = json.load(json_file)
		except Exception as e:
			print("")
			print(f"ERROR: parsing json from '{credentials_fn}':")
			print("")
			print(f"	   {e}")
			print("")
			return None
		credentials = merge(default_credentials, loaded_credentials)

	if credentials["printful-api-key"] == default_credentials["printful-api-key"]:
		print(f"NOTICE: {credentials_fn} contains default values for some credentials\n		If there are any authentication error this might be the culprit.")

	# Printify needs printful-api-key base64 encoded, so do that now b64encode
	credentials["printful-api-key-base64"] = str(base64.urlsafe_b64encode(credentials["printful-api-key"].encode("utf-8")).strip(), "utf-8")

	# Print credentials to terminal for reference
	# print('Credentials: ') 	pprint.pprint(credentials)
	return credentials
